"""Classes and functions for working with TextGrids in MFA"""
from __future__ import annotations

import os
import re
import sys
import traceback
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Union

from praatio import textgrid as tgio
from praatio.utilities.textgrid_io import Interval

if TYPE_CHECKING:
    from .aligner.base import BaseAligner
    from .corpus.classes import DictionaryData, File, Speaker
    from .dictionary import (
        DictionaryEntryType,
        IpaType,
        MappingType,
        PunctuationType,
        ReversedMappingType,
    )
    from .multiprocessing.alignment import CtmType
    from .trainers.base import BaseTrainer

    AlignerType = Union[BaseTrainer, BaseAligner]

__all__ = [
    "CtmInterval",
    "process_ctm_line",
    "map_to_original_pronunciation",
    "parse_from_word",
    "parse_from_phone",
    "parse_from_word_no_cleanup",
    "generate_tiers",
    "export_textgrid",
    "ctm_to_textgrid",
    "output_textgrid_writing_errors",
    "to_int",
    "ctms_to_textgrids_non_mp",
    "split_clitics",
]


@dataclass
class CtmInterval:
    """
    Data class for intervals derived from CTM files

    Attributes
    ----------
    begin: float
        Start time of interval
    end: float
        End time of interval
    label: str
        Text of interval
    utterance: str
        Utterance ID that the interval belongs to
    """

    begin: float
    end: float
    label: str
    utterance: str

    def shift_times(self, offset: float):
        """
        Shift times of the interval based on some offset (i.e., segments in Kaldi)

        Parameters
        ----------
        offset: float
            Offset to add to the interval's begin and end

        """
        self.begin += offset
        self.end += offset

    def to_tg_interval(self) -> Interval:
        """
        Converts the CTMInterval to PraatIO's Interval class

        Returns
        -------
        Interval
            Derived PraatIO Interval
        """
        return Interval(self.begin, self.end, self.label)


def process_ctm_line(line: str) -> CtmInterval:
    """
    Helper function for parsing a line of CTM file to construct a CTMInterval

    Parameters
    ----------
    line: str
        Input string

    Returns
    -------
    CtmInterval
        Extracted data from the line
    """
    line = line.split(" ")
    utt = line[0]
    begin = round(float(line[2]), 4)
    duration = float(line[3])
    end = round(begin + duration, 4)
    label = line[4]
    return CtmInterval(begin, end, label, utt)


def split_clitics(
    item: str,
    words_mapping: MappingType,
    clitic_set: Set[str],
    clitic_markers: PunctuationType,
    compound_markers: PunctuationType,
) -> List[str]:
    """
    Split a word into subwords based on dictionary information

    Parameters
    ----------
    item: str
        Word to split
    words_mapping: Dict[str, int]
        A word mapping from a Dictionary object
    clitic_set: Set[str]
        Set of acceptable clitics from the dictionary
    clitic_markers: str
        Clitic markers
    compound_markers: str
        Compound markers

    Returns
    -------
    List[str]
        List of subwords
    """
    if item in words_mapping:
        return [item]
    if any(x in item for x in compound_markers):
        s = re.split(rf"[{compound_markers}]", item)
        if any(x in item for x in clitic_markers):
            new_s = []
            for seg in s:
                if any(x in seg for x in clitic_markers):
                    new_s.extend(
                        split_clitics(
                            seg, words_mapping, clitic_set, clitic_markers, compound_markers
                        )
                    )
                else:
                    new_s.append(seg)
            s = new_s
        return s
    if any(x in item and not item.endswith(x) and not item.startswith(x) for x in clitic_markers):
        initial, final = re.split(rf"[{clitic_markers}]", item, maxsplit=1)
        if any(x in final for x in clitic_markers):
            final = split_clitics(
                final, words_mapping, clitic_set, clitic_markers, compound_markers
            )
        else:
            final = [final]
        for clitic in clitic_markers:
            if initial + clitic in clitic_set:
                return [initial + clitic] + final
            elif clitic + final[0] in clitic_set:
                final[0] = clitic + final[0]
                return [initial] + final
    return [item]


def _lookup(
    item: str,
    words_mapping: MappingType,
    punctuation: PunctuationType,
    clitic_set: Set[str],
    clitic_markers: PunctuationType,
    compound_markers: PunctuationType,
) -> List[str]:
    """
    Look up a word and return the list of sub words if necessary
    taking into account clitic and compound markers

    Parameters
    ----------
    item: str
        Word to look up

    Returns
    -------
    List[str]
        List of subwords that are in the dictionary
    """
    from montreal_forced_aligner.dictionary import sanitize

    if item in words_mapping:
        return [item]
    sanitized = sanitize(item, punctuation, clitic_markers)
    if sanitized in words_mapping:
        return [sanitized]
    split = split_clitics(sanitized, words_mapping, clitic_set, clitic_markers, compound_markers)
    oov_count = sum(1 for x in split if x not in words_mapping)

    if oov_count < len(split):  # Only returned split item if it gains us any transcribed speech
        return split
    return [sanitized]


def to_int(
    item: str,
    words_mapping: MappingType,
    punctuation: PunctuationType,
    clitic_set: Set[str],
    clitic_markers: PunctuationType,
    compound_markers: PunctuationType,
    oov_int: int,
) -> List[int]:
    """
    Convert a given word into integer IDs

    Parameters
    ----------
    item: str
        Word to look up

    Returns
    -------
    List[int]
        List of integer IDs corresponding to each subword
    """
    if item == "":
        return []
    sanitized = _lookup(
        item, words_mapping, punctuation, clitic_set, clitic_markers, compound_markers
    )
    text_int = []
    for item in sanitized:
        if not item:
            continue
        if item not in words_mapping:
            text_int.append(oov_int)
        else:
            text_int.append(words_mapping[item])
    return text_int


def parse_from_word(
    ctm_labels: List[CtmInterval], text: List[str], dictionary_data: DictionaryData
) -> List[CtmInterval]:
    """
    Parse CTM intervals into the corresponding text for an utterance

    Parameters
    ----------
    ctm_labels: List[CtmInterval]
        CTM intervals
    text: List[str]
        The original text that was to be aligned
    dictionary_data: DictionaryData
        Dictionary data necessary for splitting subwords

    Returns
    -------
    List[CtmInterval]
        Correct intervals with subwords merged back into their original text
    """
    cur_ind = 0
    actual_labels = []
    utterance = None
    words_mapping = dictionary_data.words_mapping
    punctuation = dictionary_data.punctuation
    clitic_set = dictionary_data.clitic_set
    clitic_markers = dictionary_data.clitic_markers
    compound_markers = dictionary_data.compound_markers
    oov_int = dictionary_data.oov_int
    for word in text:
        ints = to_int(
            word, words_mapping, punctuation, clitic_set, clitic_markers, compound_markers, oov_int
        )
        b = 1000000
        e = -1
        for i in ints:
            cur = ctm_labels[cur_ind]
            if utterance is None:
                utterance = cur.utterance
            if i == int(cur.label):
                if cur.begin < b:
                    b = cur.begin
                if cur.end > e:
                    e = cur.end
            cur_ind += 1
        lab = CtmInterval(b, e, word, utterance)
        actual_labels.append(lab)
    return actual_labels


def parse_from_word_no_cleanup(
    ctm_labels: List[CtmInterval], reversed_word_mapping: ReversedMappingType
) -> List[CtmInterval]:
    """
    Assume that subwords in the CTM files are desired, so just does a reverse look up to get the sub word
    text

    Parameters
    ----------
    ctm_labels: List[CtmInterval]
        List of CtmIntervals to convert
    reversed_word_mapping: Dict[int, str]
        Look up for Kaldi word IDs to convert them back to text

    Returns
    -------
    List[CtmInterval]
        Parsed intervals with text rather than integer IDs
    """
    for ctm_interval in ctm_labels:
        label = reversed_word_mapping[int(ctm_interval.label)]
        ctm_interval.label = label
    return ctm_labels


def parse_from_phone(
    ctm_labels: List[CtmInterval],
    reversed_phone_mapping: ReversedMappingType,
    positions: List[str],
) -> List[CtmInterval]:
    """
    Parse CtmIntervals to original phone transcriptions

    Parameters
    ----------
    ctm_labels: List[CtmInterval]
        CtmIntervals to convert
    reversed_phone_mapping: Dict[int, str]
        Mapping to convert phone IDs to phone labels
    positions: List[str]
        List of word positions to account for

    Returns
    -------
    List[CtmInterval]
        Parsed intervals with phone labels rather than IDs
    """
    for ctm_interval in ctm_labels:
        label = reversed_phone_mapping[int(ctm_interval.label)]
        for p in positions:
            if label.endswith(p):
                label = label[: -1 * len(p)]
        ctm_interval.label = label
    return ctm_labels


def map_to_original_pronunciation(
    phones: CtmType, subpronunciations: List[DictionaryEntryType], strip_diacritics: IpaType
) -> CtmType:
    """
    Convert phone transcriptions from multilingual IPA mode to their original IPA transcription

    Parameters
    ----------
    phones: List[CtmInterval]
        List of aligned phones
    subpronunciations: List[DictionaryEntryType]
        Pronunciations of each sub word to reconstruct the transcriptions
    strip_diacritics: List[str]
        List of diacritics that were stripped out of the original IPA transcription

    Returns
    -------
    List[CtmInterval]
        Intervals with their original IPA pronunciation rather than the internal simplified form
    """
    transcription = tuple(x.label for x in phones)
    new_phones = []
    mapping_ind = 0
    transcription_ind = 0
    for pronunciations in subpronunciations:
        pron = None
        if mapping_ind >= len(phones):
            break
        for p in pronunciations:
            if (
                "original_pronunciation" in p
                and transcription == p["pronunciation"] == p["original_pronunciation"]
            ) or (transcription == p["pronunciation"] and "original_pronunciation" not in p):
                new_phones.extend(phones)
                mapping_ind += len(phones)
                break
            if (
                p["pronunciation"]
                == transcription[transcription_ind : transcription_ind + len(p["pronunciation"])]
                and pron is None
            ):
                pron = p
        if mapping_ind >= len(phones):
            break
        if not pron:
            new_phones.extend(phones)
            mapping_ind += len(phones)
            break
        to_extend = phones[transcription_ind : transcription_ind + len(pron["pronunciation"])]
        transcription_ind += len(pron["pronunciation"])
        p = pron
        if "original_pronunciation" not in p or p["pronunciation"] == p["original_pronunciation"]:

            new_phones.extend(to_extend)
            mapping_ind += len(to_extend)
            break
        for pi in p["original_pronunciation"]:
            if pi == phones[mapping_ind].label:
                new_phones.append(phones[mapping_ind])
            else:
                modded_phone = pi
                new_p = phones[mapping_ind].label
                for diacritic in strip_diacritics:
                    modded_phone = modded_phone.replace(diacritic, "")
                if modded_phone == new_p:
                    phones[mapping_ind].label = pi
                    new_phones.append(phones[mapping_ind])
                elif mapping_ind != len(phones) - 1:
                    new_p = phones[mapping_ind].label + phones[mapping_ind + 1].label
                    if modded_phone == new_p:
                        new_phones.append(
                            CtmInterval(
                                phones[mapping_ind].begin,
                                phones[mapping_ind + 1].end,
                                new_p,
                                phones[mapping_ind].utterance,
                            )
                        )
                        mapping_ind += 1
            mapping_ind += 1
    return new_phones


def ctms_to_textgrids_non_mp(aligner: AlignerType) -> None:
    """
    Parse CTM files to TextGrids without using multiprocessing

    Parameters
    ----------
    aligner: AlignerType
        Aligner that generated the CTM files
    """

    def process_current_word_labels():
        """Process the current stack of word labels"""
        speaker = cur_utt.speaker

        text = cur_utt.text.split()
        if aligner.align_config.cleanup_textgrids:
            actual_labels = parse_from_word(current_labels, text, speaker.dictionary_data)
        else:
            actual_labels = parse_from_word_no_cleanup(
                current_labels, speaker.dictionary_data.reversed_words_mapping
            )
        cur_utt.word_labels = actual_labels

    def process_current_phone_labels():
        """Process the current stack of phone labels"""
        speaker = cur_utt.speaker

        cur_utt.phone_labels = parse_from_phone(
            current_labels, speaker.dictionary.reversed_phone_mapping, speaker.dictionary.positions
        )

    export_errors = {}
    for j in aligner.corpus.jobs:

        word_arguments = j.cleanup_word_ctm_arguments(aligner)
        phone_arguments = j.phone_ctm_arguments(aligner)
        aligner.logger.debug(f"Parsing ctms for job {j.name}...")
        cur_utt = None
        current_labels = []
        for dict_name in word_arguments.dictionaries:
            with open(word_arguments.ctm_paths[dict_name], "r") as f:
                for line in f:
                    line = line.strip()
                    if line == "":
                        continue
                    ctm_interval = process_ctm_line(line)
                    utt = aligner.corpus.utterances[ctm_interval.utterance]
                    if cur_utt is None:
                        cur_utt = utt
                    if utt.is_segment:
                        utt_begin = utt.begin
                    else:
                        utt_begin = 0
                    if utt != cur_utt:
                        process_current_word_labels()
                        cur_utt = utt
                        current_labels = []

                    ctm_interval.shift_times(utt_begin)
                    current_labels.append(ctm_interval)
            if current_labels:
                process_current_word_labels()
        cur_utt = None
        current_labels = []
        for dict_name in phone_arguments.dictionaries:
            with open(phone_arguments.ctm_paths[dict_name], "r") as f:
                for line in f:
                    line = line.strip()
                    if line == "":
                        continue
                    ctm_interval = process_ctm_line(line)
                    utt = aligner.corpus.utterances[ctm_interval.utterance]
                    if cur_utt is None:
                        cur_utt = utt
                    if utt.is_segment:
                        utt_begin = utt.begin
                    else:
                        utt_begin = 0
                    if utt != cur_utt and cur_utt is not None:
                        process_current_phone_labels()
                        cur_utt = utt
                        current_labels = []

                    ctm_interval.shift_times(utt_begin)
                    current_labels.append(ctm_interval)
            if current_labels:
                process_current_phone_labels()

        aligner.logger.debug(f"Generating TextGrids for job {j.name}...")
        processed_files = set()
        for file in j.job_files().values():
            first_file_write = True
            if file.name in processed_files:
                first_file_write = False
            try:
                ctm_to_textgrid(file, aligner, first_file_write)
                processed_files.add(file.name)
            except Exception:
                if aligner.align_config.debug:
                    raise
                exc_type, exc_value, exc_traceback = sys.exc_info()
                export_errors[file.name] = "\n".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )
    if export_errors:
        aligner.logger.warning(
            f"There were {len(export_errors)} errors encountered in generating TextGrids. "
            f"Check the output_errors.txt file in {os.path.join(aligner.textgrid_output)} "
            f"for more details"
        )
    output_textgrid_writing_errors(aligner.textgrid_output, export_errors)


def output_textgrid_writing_errors(output_directory: str, export_errors: Dict[str, str]) -> None:
    """
    Output any errors that were encountered in writing TextGrids

    Parameters
    ----------
    output_directory: str
        Directory to save TextGrids files
    export_errors: Dict[str, str]
        Dictionary of errors encountered
    """
    error_log = os.path.join(output_directory, "output_errors.txt")
    if os.path.exists(error_log):
        os.remove(error_log)
    for file_name, result in export_errors.items():
        if not os.path.exists(error_log):
            with open(error_log, "w", encoding="utf8") as f:
                f.write(
                    "The following exceptions were encountered during the output of the alignments to TextGrids:\n\n"
                )
        with open(error_log, "a", encoding="utf8") as f:
            f.write(f"{file_name}:\n")
            f.write(f"{result}\n\n")


def generate_tiers(
    file: File, cleanup_textgrids: Optional[bool] = True
) -> Dict[Speaker, Dict[str, CtmType]]:
    """
    Generate TextGrid tiers for a given File

    Parameters
    ----------
    file: File
        File to generate TextGrid tiers
    cleanup_textgrids: bool, optional
        Flag for whether the TextGrids should be cleaned up

    Returns
    -------
    Dict[Speaker, Dict[str, CtmType]]
        Tier information per speaker, with CtmIntervals split by "phones" and "words"
    """
    output = {}

    for u in file.utterances.values():
        if not u.word_labels:
            continue
        speaker = u.speaker
        dictionary_data = speaker.dictionary_data

        words = []
        phones = []
        if dictionary_data.multilingual_ipa and cleanup_textgrids:
            phone_ind = 0
            for interval in u.word_labels:
                end = interval.end
                word = interval.label
                subwords = _lookup(
                    word,
                    dictionary_data.words_mapping,
                    dictionary_data.punctuation,
                    dictionary_data.clitic_set,
                    dictionary_data.clitic_markers,
                    dictionary_data.compound_markers,
                )
                subwords = [
                    x if x in dictionary_data.words_mapping else dictionary_data.oov_code
                    for x in subwords
                ]
                subprons = [dictionary_data.words[x] for x in subwords]
                cur_phones = []
                while u.phone_labels[phone_ind].end <= end:
                    p = u.phone_labels[phone_ind]
                    if p.label in dictionary_data.silences:
                        phone_ind += 1
                        continue
                    cur_phones.append(p)
                    phone_ind += 1
                    if phone_ind > len(u.phone_labels) - 1:
                        break
                phones.extend(
                    map_to_original_pronunciation(
                        cur_phones, subprons, dictionary_data.strip_diacritics
                    )
                )
                if not word:
                    continue

                words.append(interval)
        else:
            for interval in u.word_labels:
                words.append(interval)
            for interval in u.phone_labels:
                if interval.label in dictionary_data.silences and cleanup_textgrids:
                    continue
                phones.append(interval)
        if speaker not in output:
            output[speaker] = {"words": words, "phones": phones}
        else:
            output[speaker]["words"].extend(words)
            output[speaker]["phones"].extend(phones)
    return output


def export_textgrid(
    file: File,
    output_path: str,
    speaker_data: Dict[Speaker, Dict[str, CtmType]],
    frame_shift: int,
    first_file_write: Optional[bool] = True,
) -> None:
    """
    Export aligned file to TextGrid

    Parameters
    ----------
    file: File
        File object to export
    output_path: str
        Output path of the file
    speaker_data: Dict[Speaker, Dict[str, CtmType]]
        Per speaker, per word/phone CtmIntervals
    frame_shift: int
        Frame shift of features, in ms
    first_file_write: bool, optional
        Flag for whether the file should be created from scratch or appended to if it
        has been modified by another export process
    """
    if frame_shift > 1:
        frame_shift = round(frame_shift / 1000, 4)
    if first_file_write:
        # Create initial textgrid
        tg = tgio.Textgrid()
        tg.minTimestamp = 0
        tg.maxTimestamp = file.duration

        if len(file.speaker_ordering) > 1:
            for speaker in file.speaker_ordering:
                word_tier_name = f"{speaker} - words"
                phone_tier_name = f"{speaker} - phones"

                word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
                phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
                tg.addTier(word_tier)
                tg.addTier(phone_tier)
        else:
            word_tier_name = "words"
            phone_tier_name = "phones"
            word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
            phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
            tg.addTier(word_tier)
            tg.addTier(phone_tier)
    else:
        # Use existing
        tg = tgio.openTextgrid(output_path, includeEmptyIntervals=False)

        word_tier_name = "words"
        phone_tier_name = "phones"
        word_tier = tgio.IntervalTier(word_tier_name, [], minT=0, maxT=file.duration)
        phone_tier = tgio.IntervalTier(phone_tier_name, [], minT=0, maxT=file.duration)
        tg.addTier(word_tier)
        tg.addTier(phone_tier)
    for speaker, data in speaker_data.items():
        words = data["words"]
        phones = data["phones"]
        tg_words = []
        tg_phones = []
        for w in words:
            if file.duration - w.end < frame_shift:  # Fix rounding issues
                w.end = file.duration
            tg_words.append(w.to_tg_interval())
        for p in phones:
            if file.duration - p.end < frame_shift:  # Fix rounding issues
                p.end = file.duration
            tg_phones.append(p.to_tg_interval())

        if len(file.speaker_ordering) > 1:
            word_tier_name = f"{speaker} - words"
            phone_tier_name = f"{speaker} - phones"
        else:
            word_tier_name = "words"
            phone_tier_name = "phones"
        word_tier = tgio.IntervalTier(word_tier_name, tg_words, minT=0, maxT=file.duration)
        phone_tier = tgio.IntervalTier(phone_tier_name, tg_phones, minT=0, maxT=file.duration)
        tg.replaceTier(word_tier_name, word_tier)
        tg.replaceTier(phone_tier_name, phone_tier)

    tg.save(output_path, includeBlankSpaces=True, format="long_textgrid", reportingMode="error")


def ctm_to_textgrid(file: File, aligner: AlignerType, first_file_write=True) -> None:
    """
    Export a File to TextGrid

    Parameters
    ----------
    file: File
        File to export
    aligner: AlignerType
        Aligner used to generate the alignments
    first_file_write: bool, optional
        Flag for whether this is the first time touching this file
    """
    data = generate_tiers(file, cleanup_textgrids=aligner.align_config.cleanup_textgrids)

    backup_output_directory = None
    if not aligner.align_config.overwrite:
        backup_output_directory = aligner.backup_output_directory
        os.makedirs(backup_output_directory, exist_ok=True)
    output_path = file.construct_output_path(aligner.textgrid_output, backup_output_directory)
    export_textgrid(
        file, output_path, data, aligner.align_config.feature_config.frame_shift, first_file_write
    )



"""Class definition for BaseTrainer"""
from __future__ import annotations

import os
import re
import shutil
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from tqdm import tqdm

from ..config import FeatureConfig
from ..exceptions import KaldiProcessingError, TrainerError
from ..models import AcousticModel
from ..multiprocessing import (
    acc_stats,
    align,
    compile_information,
    compile_train_graphs,
    compute_alignment_improvement,
    convert_ali_to_textgrids,
)
from ..utils import log_kaldi_errors, parse_logs

if TYPE_CHECKING:
    from ..config import ConfigDict
    from ..corpus import Corpus
    from ..dictionary import DictionaryType
    from ..models import MetaDict


__all__ = ["BaseTrainer"]


class BaseTrainer:
    """
    Base trainer class for training acoustic models and ivector extractors

    Attributes
    ----------
    num_iterations : int
        Number of training iterations to perform, defaults to 40
    transition_scale : float
        Scaling of transition costs in alignment, defaults to 1.0
    acoustic_scale : float
        Scaling of acoustic costs in alignment, defaults to 0.1
    self_loop_scale : float
        Scaling of self loop costs in alignment, defaults to 0.1
    beam : int
        Default beam width for alignment, defaults = 10
    retry_beam : int
        Beam width to fall back on if no alignment is produced, defaults to 40
    max_gaussians : int
        Total number of gaussians, defaults to 1000
    boost_silence : float
        Factor by which to boost silence likelihoods in alignment, defaults to 1.0
    realignment_iterations : list
        List of iterations to perform alignment
    power : float
        Exponent for number of gaussians according to occurrence counts, defaults to 0.25
    debug: bool
        Flag for debug mode
    use_mp: bool
        Flag for whether to use multiprocessing
    iteration: int
        Current iteration
    training_complete: bool
        Flag for whether training has been successfully completed
    speaker_independent: bool
        Flag for using speaker-independent features regardless of speaker adaptation
    uses_cmvn: bool
        Flag for whether to include CMVN in features
    uses_splices: bool
        Flag for whether to include splices in features
    uses_voiced: bool
        Flag for whether to use voiced features
    """

    def __init__(self, default_feature_config: FeatureConfig):
        self.logger = None
        self.dictionary = None
        self.transition_scale = 1.0
        self.acoustic_scale = 0.1
        self.self_loop_scale = 0.1
        self.realignment_iterations = []
        self.num_iterations = 40
        self.beam = 10
        self.retry_beam = 40
        self.max_gaussians = 1000
        self.boost_silence = 1.0
        self.power = 0.25
        self.subset = None
        self.calc_pron_probs = False
        self.architecture = "gmm-hmm"
        self.feature_config = FeatureConfig()
        self.feature_config.update(default_feature_config.params())
        self.initial_gaussians = None  # Gets set later
        self.temp_directory = None
        self.identifier = None
        self.corpus: Optional[Corpus] = None
        self.data_directory = None
        self.debug = False
        self.use_mp = True
        self.current_gaussians = None
        self.iteration = 0
        self.training_complete = False
        self.speaker_independent = True
        self.uses_cmvn = True
        self.uses_splices = False
        self.uses_voiced = False
        self.previous_trainer: Optional[BaseTrainer] = None

    @property
    def train_directory(self) -> str:
        """Training directory"""
        return os.path.join(self.temp_directory, self.identifier)

    @property
    def log_directory(self) -> str:
        """Training log directory"""
        return os.path.join(self.train_directory, "log")

    @property
    def align_directory(self) -> str:
        """Alignment directory"""
        return os.path.join(self.temp_directory, f"{self.identifier}_ali")

    @property
    def align_log_directory(self) -> str:
        """Alignment log directory"""
        return os.path.join(self.align_directory, "log")

    @property
    def working_directory(self) -> str:
        """Current working directory"""
        if self.training_complete:
            return self.align_directory
        return self.train_directory

    @property
    def working_log_directory(self) -> str:
        """Log directory of current working directory"""
        if self.training_complete:
            return self.align_log_directory
        return self.log_directory

    @property
    def fmllr_options(self) -> ConfigDict:
        """Options for fMLLR calculation, only used by SatTrainer"""
        raise NotImplementedError

    @property
    def lda_options(self) -> ConfigDict:
        """Options for LDA calculation, only used by LdaTrainer"""
        raise NotImplementedError

    @property
    def tree_path(self):
        """Path to tree file"""
        return os.path.join(self.working_directory, "tree")

    @property
    def current_model_path(self):
        """Current acoustic model path"""
        if (
            self.training_complete
            or self.iteration is None
            or self.iteration > self.num_iterations
        ):
            return os.path.join(self.working_directory, "final.mdl")
        return os.path.join(self.working_directory, f"{self.iteration}.mdl")

    @property
    def next_model_path(self):
        """Next iteration's acoustic model path"""
        if self.iteration > self.num_iterations:
            return os.path.join(self.working_directory, "final.mdl")
        return os.path.join(self.working_directory, f"{self.iteration + 1}.mdl")

    @property
    def next_occs_path(self):
        """Next iteration's occs file path"""
        if self.training_complete:
            return os.path.join(self.working_directory, "final.occs")
        return os.path.join(self.working_directory, f"{self.iteration + 1}.occs")

    @property
    def alignment_model_path(self):
        """Alignment model path"""
        path = os.path.join(self.working_directory, "final.alimdl")
        if self.speaker_independent and os.path.exists(path):
            return path
        if not self.training_complete:
            return self.current_model_path
        return os.path.join(self.working_directory, "final.mdl")

    def compute_calculated_properties(self) -> None:
        """Compute any calculated properties such as alignment iterations"""
        pass

    @property
    def train_type(self) -> str:
        """Training type, not implemented for BaseTrainer"""
        raise NotImplementedError

    @property
    def phone_type(self) -> str:
        """Phone type, not implemented for BaseTrainer"""
        raise NotImplementedError

    @property
    def final_gaussian_iteration(self) -> int:
        """Final iteration to increase gaussians"""
        return self.num_iterations - 10

    @property
    def gaussian_increment(self) -> int:
        """Amount by which gaussians should be increases each iteration"""
        return int((self.max_gaussians - self.initial_gaussians) / self.final_gaussian_iteration)

    @property
    def align_options(self) -> ConfigDict:
        """Options for alignment"""
        options_silence_csl = ""
        if self.dictionary:
            options_silence_csl = self.dictionary.optional_silence_csl
        return {
            "beam": self.beam,
            "retry_beam": self.retry_beam,
            "transition_scale": self.transition_scale,
            "acoustic_scale": self.acoustic_scale,
            "self_loop_scale": self.self_loop_scale,
            "boost_silence": self.boost_silence,
            "debug": self.debug,
            "optional_silence_csl": options_silence_csl,
        }

    def analyze_align_stats(self) -> None:
        """
        Analyzes alignment stats and outputs debug information
        """
        unaligned, log_like = compile_information(self)

        self.logger.debug(
            f"Average per frame likelihood (this might not actually mean anything) "
            f"for {self.identifier}: {log_like}"
        )
        self.logger.debug(f"Number of unaligned files " f"for {self.identifier}: {len(unaligned)}")

    def update(self, data: Dict[str, Any]) -> None:
        """
        Update configuration data

        Parameters
        ----------
        data: Dict[str, Any]
            Data to update
        """
        from ..config.base_config import PARSING_KEYS

        for k, v in data.items():
            if k == "use_mp":
                self.feature_config.use_mp = v
            if k == "features":
                self.feature_config.update(v)
            elif k in PARSING_KEYS:
                continue
            elif not hasattr(self, k):
                raise TrainerError(f"No field found for key {k}")
            else:
                setattr(self, k, v)
        self.compute_calculated_properties()

    def _setup_for_init(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[BaseTrainer],
    ) -> None:
        """
        Default initialization for all Trainers

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: :class:`~montreal_forced_aligner.trainers.base.BaseTrainer`, optional
            Previous trainer to initialize from

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        begin = time.time()
        self.temp_directory = temporary_directory
        self.identifier = identifier
        dirty_path = os.path.join(self.train_directory, "dirty")
        done_path = os.path.join(self.align_directory, "done")
        if os.path.exists(dirty_path):  # if there was an error, let's redo from scratch
            shutil.rmtree(self.train_directory)
        self.logger.info(f"Initializing training for {identifier}...")
        self.corpus = corpus
        try:
            self.data_directory = self.corpus.split_directory
            self.corpus.generate_features()
            if self.subset is not None:
                self.data_directory = self.corpus.subset_directory(self.subset)
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        self.dictionary = dictionary
        self.previous_trainer = previous_trainer
        if os.path.exists(done_path):
            self.training_complete = True
            self.iteration = None
            return
        os.makedirs(self.train_directory, exist_ok=True)
        os.makedirs(self.log_directory, exist_ok=True)
        if self.subset is not None and self.subset > corpus.num_utterances:
            self.logger.warning(
                "Subset specified is larger than the dataset, "
                "using full corpus for this training block."
            )

        self.logger.debug(f"Setup for initialization took {time.time() - begin} seconds")

    def increment_gaussians(self):
        """Increment the current number of gaussians"""
        self.current_gaussians += self.gaussian_increment

    def init_training(
        self,
        identifier: str,
        temporary_directory: str,
        corpus: Corpus,
        dictionary: DictionaryType,
        previous_trainer: Optional[BaseTrainer],
    ) -> None:
        """
        Initialize training, not implemented for BaseTrainer

        Parameters
        ----------
        identifier: str
            Identifier for the training block
        temporary_directory: str
            Root temporary directory to save
        corpus: :class:`~montreal_forced_aligner.corpus.base.Corpus`
            Corpus to use
        dictionary: DictionaryType
            Dictionary to use
        previous_trainer: :class:`~montreal_forced_aligner.trainers.base.BaseTrainer`, optional
            Previous trainer to initialize from
        """
        raise NotImplementedError

    def get_unaligned_utterances(self) -> List[str]:
        """Find all utterances that were not aligned for validation utility"""
        error_regex = re.compile(r"Did not successfully decode file (\w+),")
        error_files = []
        for j in self.corpus.jobs:
            path = os.path.join(self.align_directory, "log", f"align.{j.name}.log")
            if not os.path.exists(path):
                continue
            with open(path, "r") as f:
                error_files.extend(error_regex.findall(f.read()))
        return error_files

    def align(self, subset: Optional[int] = None) -> None:
        """
        Align a subset of the corpus for the next trainer

        Parameters
        ----------
        subset: int, optional
            Number of utterances to include in the subset

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        if not os.path.exists(self.align_directory):
            self.finalize_training()
        dirty_path = os.path.join(self.align_directory, "dirty")
        if os.path.exists(dirty_path):  # if there was an error, let's redo from scratch
            shutil.rmtree(self.align_directory)
        done_path = os.path.join(self.align_directory, "done")
        if not os.path.exists(done_path):
            message = f"Generating alignments using {self.identifier} models"
            if subset:
                message += f" using {subset} utterances..."
            else:
                message += " for the whole corpus..."
            self.logger.info(message)
            begin = time.time()
            if subset is None:
                self.data_directory = self.corpus.split_directory
            else:
                self.data_directory = self.corpus.subset_directory(subset)
            try:
                self.iteration = None
                compile_train_graphs(self)
                align(self)
                self.analyze_align_stats()
                self.save(os.path.join(self.align_directory, "acoustic_model.zip"))
            except Exception as e:
                with open(dirty_path, "w"):
                    pass
                if isinstance(e, KaldiProcessingError):
                    log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
                raise
            with open(done_path, "w"):
                pass
            self.logger.debug(f"Alignment took {time.time() - begin} seconds")
        else:
            self.logger.info(f"Alignments using {self.identifier} models already done")

    def training_iteration(self):
        """Perform an iteration of training"""
        if os.path.exists(self.next_model_path):
            self.iteration += 1
            return
        if self.iteration in self.realignment_iterations:
            align(self)
            if self.debug:
                compute_alignment_improvement(self)
        acc_stats(self)

        parse_logs(self.log_directory)
        if self.iteration < self.final_gaussian_iteration:
            self.increment_gaussians()
        self.iteration += 1

    def train(self):
        """
        Train the model

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        done_path = os.path.join(self.train_directory, "done")
        dirty_path = os.path.join(self.train_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info(f"{self.identifier} training already done, skipping initialization.")
            return
        begin = time.time()
        try:
            with tqdm(total=self.num_iterations) as pbar:
                while self.iteration < self.num_iterations + 1:
                    self.training_iteration()
                    pbar.update(1)
            self.finalize_training()
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        with open(done_path, "w"):
            pass
        self.logger.info("Training complete!")
        self.logger.debug(f"Training took {time.time() - begin} seconds")

    def finalize_training(self):
        """
        Finalize the training, moving all relevant files from the training directory to the
        alignment directory and changing flags to point at align directory as the working directory

        """
        os.makedirs(self.align_directory, exist_ok=True)
        os.makedirs(self.align_log_directory, exist_ok=True)
        shutil.copy(
            os.path.join(self.train_directory, f"{self.num_iterations}.mdl"),
            os.path.join(self.train_directory, "final.mdl"),
        )
        shutil.copy(
            os.path.join(self.train_directory, f"{self.num_iterations}.occs"),
            os.path.join(self.train_directory, "final.occs"),
        )
        shutil.copy(os.path.join(self.train_directory, "tree"), self.align_directory)
        shutil.copyfile(
            os.path.join(self.train_directory, "final.mdl"),
            os.path.join(self.align_directory, "final.mdl"),
        )

        if os.path.exists(os.path.join(self.train_directory, "lda.mat")):
            shutil.copyfile(
                os.path.join(self.train_directory, "lda.mat"),
                os.path.join(self.align_directory, "lda.mat"),
            )
        shutil.copyfile(
            os.path.join(self.train_directory, "final.occs"),
            os.path.join(self.align_directory, "final.occs"),
        )
        if not self.debug:
            for i in range(1, self.num_iterations):
                model_path = os.path.join(self.train_directory, f"{i}.mdl")
                try:
                    os.remove(model_path)
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(self.train_directory, f"{i}.occs"))
                except FileNotFoundError:
                    pass
        self.training_complete = True
        self.iteration = None

    @property
    def meta(self) -> MetaDict:
        """Generate metadata for the acoustic model that was trained"""
        from datetime import datetime

        from ..utils import get_mfa_version

        data = {
            "phones": sorted(self.dictionary.nonsil_phones),
            "version": get_mfa_version(),
            "architecture": self.architecture,
            "train_date": str(datetime.now()),
            "features": self.feature_config.params(),
            "multilingual_ipa": self.dictionary.multilingual_ipa,
        }
        if self.dictionary.multilingual_ipa:
            data["strip_diacritics"] = self.dictionary.strip_diacritics
            data["digraphs"] = self.dictionary.digraphs
        return data

    def export_textgrids(self) -> None:
        """
        Export a TextGrid file for every sound file in the dataset

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        begin = time.time()
        try:
            convert_ali_to_textgrids(self)
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise
        self.logger.debug(f"Exporting textgrids took {time.time() - begin} seconds")

    def save(self, path: str, root_directory: Optional[str] = None) -> None:
        """
        Export an acoustic model and dictionary to the specified path

        Parameters
        ----------
        path : str
            Path to save acoustic model and dictionary
        root_directory : str or None
            Path for root directory of temporary files
        """
        directory, filename = os.path.split(path)
        basename, _ = os.path.splitext(filename)
        acoustic_model = AcousticModel.empty(basename, root_directory=root_directory)
        acoustic_model.add_meta_file(self)
        acoustic_model.add_model(self.train_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(path)
        acoustic_model.dump(path)



"""Class definitions for configuring aligning"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Collection

import yaml

from .base_config import (
    DEFAULT_CLITIC_MARKERS,
    DEFAULT_COMPOUND_MARKERS,
    DEFAULT_DIGRAPHS,
    DEFAULT_PUNCTUATION,
    DEFAULT_STRIP_DIACRITICS,
    BaseConfig,
    ConfigError,
)
from .feature_config import FeatureConfig

if TYPE_CHECKING:
    from argparse import Namespace

    from . import ConfigDict

__all__ = ["AlignConfig", "align_yaml_to_config", "load_basic_align"]


class AlignConfig(BaseConfig):
    """
    Configuration object for alignment

    Attributes
    ----------
    transition_scale : float
        Transition scale, defaults to 1.0
    acoustic_scale : float
        Acoustic scale, defaults to 0.1
    self_loop_scale : float
        Self-loop scale, defaults to 0.1
    disable_sat : bool
        Flag for disabling speaker adaptation, defaults to False
    feature_config : :class:`~montreal_forced_aligner.features.config.FeatureConfig`
        Configuration object for feature generation
    boost_silence : float
        Factor to boost silence probabilities, 1.0 is no boost or reduction
    beam : int
        Size of the beam to use in decoding, defaults to 10
    retry_beam : int
        Size of the beam to use in decoding if it fails with the initial beam width, defaults to 40
    data_directory : str
        Path to save feature files
    fmllr_update_type : str
        Type of update for fMLLR, defaults to full
    use_mp : bool
        Flag for whether to use multiprocessing in feature generation
    """

    def __init__(self, feature_config: FeatureConfig):
        self.transition_scale = 1.0
        self.acoustic_scale = 0.1
        self.self_loop_scale = 0.1
        self.disable_sat = False
        self.feature_config = feature_config
        self.boost_silence = 1.0
        self.beam = 10
        self.retry_beam = 40
        self.data_directory = None  # Gets set later
        self.fmllr_update_type = "full"
        self.punctuation = DEFAULT_PUNCTUATION
        self.clitic_markers = DEFAULT_CLITIC_MARKERS
        self.compound_markers = DEFAULT_COMPOUND_MARKERS
        self.strip_diacritics = DEFAULT_STRIP_DIACRITICS
        self.digraphs = DEFAULT_DIGRAPHS
        self.use_mp = True
        self.use_fmllr_mp = False
        self.multilingual_ipa = False
        self.debug = False
        self.overwrite = False
        self.cleanup_textgrids = True
        self.initial_fmllr = True
        self.iteration = None

    @property
    def align_options(self) -> ConfigDict:
        """Options for use in aligning"""
        return {
            "transition_scale": self.transition_scale,
            "acoustic_scale": self.acoustic_scale,
            "self_loop_scale": self.self_loop_scale,
            "beam": self.beam,
            "retry_beam": self.retry_beam,
            "boost_silence": self.boost_silence,
            "debug": self.debug,
        }

    @property
    def fmllr_options(self) -> ConfigDict:
        """Options for use in calculating fMLLR transforms"""
        return {
            "fmllr_update_type": self.fmllr_update_type,
        }

    def update(self, data: dict) -> None:
        """Update configuration"""
        for k, v in data.items():
            if k == "use_mp":
                self.feature_config.use_mp = v
            elif k in ["punctuation", "clitic_markers", "compound_markers"]:
                if not v:
                    continue
                if "-" in v:
                    v = "-" + v.replace("-", "")
                if "]" in v and r"\]" not in v:
                    v = v.replace("]", r"\]")
            elif not hasattr(self, k):
                raise ConfigError(f"No field found for key {k}")
            setattr(self, k, v)

    def update_from_args(self, args: Namespace):
        """Update from command line arguments"""
        super(AlignConfig, self).update_from_args(args)
        self.feature_config.update_from_args(args)

    def update_from_unknown_args(self, args: Collection[str]):
        """Update from unknown command line arguments"""
        super(AlignConfig, self).update_from_unknown_args(args)
        self.feature_config.update_from_unknown_args(args)
        if self.retry_beam <= self.beam:
            self.retry_beam = self.beam * 4


def align_yaml_to_config(path: str) -> AlignConfig:
    """
    Helper function to load alignment configurations

    Parameters
    ----------
    path: str
        Path to yaml file

    Returns
    -------
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Alignment configuration
    """
    with open(path, "r", encoding="utf8") as f:
        data = yaml.load(f, Loader=yaml.SafeLoader)
        global_params = {}
        feature_config = FeatureConfig()
        for k, v in data.items():
            if k == "features":
                feature_config.update(v)
            else:
                global_params[k] = v
        align_config = AlignConfig(feature_config)
        align_config.update(global_params)
        if align_config.beam >= align_config.retry_beam:
            raise ConfigError("Retry beam must be greater than beam.")
        return align_config


def load_basic_align() -> AlignConfig:
    """
    Helper function to load the default parameters

    Returns
    -------
    :class:`~montreal_forced_aligner.config.align_config.AlignConfig`
        Default alignment configuration
    """
    base_dir = os.path.dirname(os.path.abspath(__file__))
    align_config = align_yaml_to_config(os.path.join(base_dir, "basic_align.yaml"))
    return align_config



"""Multiprocessing functions for loading corpora"""
from __future__ import annotations

import multiprocessing as mp
import sys
import traceback
from queue import Empty
from typing import TYPE_CHECKING, Dict, Union

from ..corpus.classes import parse_file
from ..exceptions import TextGridParseError, TextParseError

if TYPE_CHECKING:
    from ..corpus import OneToManyMappingType, OneToOneMappingType
    from ..corpus.base import SoundFileInfoDict

    FileInfoDict = Dict[
        str, Union[str, SoundFileInfoDict, OneToOneMappingType, OneToManyMappingType]
    ]
    from .helper import Stopped


__all__ = ["CorpusProcessWorker"]


class CorpusProcessWorker(mp.Process):
    """
    Multiprocessing corpus loading worker

    Attributes
    ----------
    job_q: :class:`~multiprocessing.Queue`
        Job queue for files to process
    return_dict: Dict
        Dictionary to catch errors
    return_q: :class:`~multiprocessing.Queue`
        Return queue for processed Files
    stopped: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Stop check for whether corpus loading should exit
    finished_adding: :class:`~montreal_forced_aligner.multiprocess.helper.Stopped`
        Signal that the main thread has stopped adding new files to be processed
    """

    def __init__(
        self,
        job_q: mp.Queue,
        return_dict: Dict,
        return_q: mp.Queue,
        stopped: Stopped,
        finished_adding: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_q = job_q
        self.return_dict = return_dict
        self.return_q = return_q
        self.stopped = stopped
        self.finished_adding = finished_adding

    def run(self) -> None:
        """
        Run the corpus loading job
        """
        while True:
            try:
                arguments = self.job_q.get(timeout=1)
            except Empty:
                if self.finished_adding.stop_check():
                    break
                continue
            self.job_q.task_done()
            if self.stopped.stop_check():
                continue
            try:
                file = parse_file(*arguments, stop_check=self.stopped.stop_check)
                self.return_q.put(file)
            except TextParseError as e:
                self.return_dict["decode_error_files"].append(e)
            except TextGridParseError as e:
                self.return_dict["textgrid_read_errors"][e.file_name] = e
            except Exception:
                self.stopped.stop()
                self.return_dict["error"] = arguments, Exception(
                    traceback.format_exception(*sys.exc_info())
                )
        return



"""Command line functions for adapting acoustic models to new data"""
from __future__ import annotations

import os
import shutil
import time
from typing import TYPE_CHECKING, Collection, Optional

from montreal_forced_aligner.aligner import AdaptingAligner, PretrainedAligner, TrainableAligner
from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.config import (
    TEMP_DIR,
    align_yaml_to_config,
    load_basic_align,
    load_command_configuration,
)
from montreal_forced_aligner.corpus import Corpus
from montreal_forced_aligner.dictionary import Dictionary, MultispeakerDictionary
from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.models import AcousticModel
from montreal_forced_aligner.utils import get_mfa_version, log_config, setup_logger

if TYPE_CHECKING:
    from argparse import Namespace

__all__ = ["adapt_model", "validate_args", "run_adapt_model"]


def adapt_model(args: Namespace, unknown_args: Optional[Collection[str]] = None) -> None:
    """
    Run the acoustic model adaptation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: List[str]
        Optional arguments that will be passed to configuration objects
    """
    command = "adapt"
    all_begin = time.time()
    if not args.temp_directory:
        temp_dir = TEMP_DIR
    else:
        temp_dir = os.path.expanduser(args.temp_directory)
    corpus_name = os.path.basename(args.corpus_directory)
    if corpus_name == "":
        args.corpus_directory = os.path.dirname(args.corpus_directory)
        corpus_name = os.path.basename(args.corpus_directory)
    data_directory = os.path.join(temp_dir, corpus_name)
    if args.config_path:
        align_config = align_yaml_to_config(args.config_path)
    else:
        align_config = load_basic_align()
    align_config.update_from_args(args)
    if unknown_args:
        align_config.update_from_unknown_args(unknown_args)
    conf_path = os.path.join(data_directory, "config.yml")
    if getattr(args, "clean", False) and os.path.exists(data_directory):
        print("Cleaning old directory!")
        shutil.rmtree(data_directory, ignore_errors=True)
    if getattr(args, "verbose", False):
        log_level = "debug"
    else:
        log_level = "info"
    logger = setup_logger(command, data_directory, console_level=log_level)
    logger.debug("ALIGN CONFIG:")
    log_config(logger, align_config)
    conf = load_command_configuration(
        conf_path,
        {
            "dirty": False,
            "begin": all_begin,
            "version": get_mfa_version(),
            "type": command,
            "corpus_directory": args.corpus_directory,
            "dictionary_path": args.dictionary_path,
            "acoustic_model_path": args.acoustic_model_path,
        },
    )
    if (
        conf["dirty"]
        or conf["type"] != command
        or conf["corpus_directory"] != args.corpus_directory
        or conf["version"] != get_mfa_version()
        or conf["dictionary_path"] != args.dictionary_path
    ):
        logger.warning(
            "WARNING: Using old temp directory, this might not be ideal for you, use the --clean flag to ensure no "
            "weird behavior for previous versions of the temporary directory."
        )
        if conf["dirty"]:
            logger.debug("Previous run ended in an error (maybe ctrl-c?)")
        if conf["type"] != command:
            logger.debug(
                f"Previous run was a different subcommand than {command} (was {conf['type']})"
            )
        if conf["corpus_directory"] != args.corpus_directory:
            logger.debug(
                "Previous run used source directory "
                f"path {conf['corpus_directory']} (new run: {args.corpus_directory})"
            )
        if conf["version"] != get_mfa_version():
            logger.debug(
                f"Previous run was on {conf['version']} version (new run: {get_mfa_version()})"
            )
        if conf["dictionary_path"] != args.dictionary_path:
            logger.debug(
                f"Previous run used dictionary path {conf['dictionary_path']} "
                f"(new run: {args.dictionary_path})"
            )
        if conf["acoustic_model_path"] != args.acoustic_model_path:
            logger.debug(
                f"Previous run used acoustic model path {conf['acoustic_model_path']} "
                f"(new run: {args.acoustic_model_path})"
            )

    os.makedirs(data_directory, exist_ok=True)
    model_directory = os.path.join(data_directory, "acoustic_models")
    os.makedirs(model_directory, exist_ok=True)
    acoustic_model = AcousticModel(args.acoustic_model_path, root_directory=model_directory)
    acoustic_model.log_details(logger)
    debug = getattr(args, "debug", False)
    audio_dir = None
    if args.audio_directory:
        audio_dir = args.audio_directory
    try:
        corpus = Corpus(
            args.corpus_directory,
            data_directory,
            speaker_characters=args.speaker_characters,
            num_jobs=args.num_jobs,
            sample_rate=align_config.feature_config.sample_frequency,
            logger=logger,
            use_mp=align_config.use_mp,
            punctuation=align_config.punctuation,
            clitic_markers=align_config.clitic_markers,
            audio_directory=audio_dir,
        )
        logger.info(corpus.speaker_utterance_info())
        if args.dictionary_path.lower().endswith(".yaml"):
            dictionary = MultispeakerDictionary(
                args.dictionary_path,
                data_directory,
                logger=logger,
                punctuation=align_config.punctuation,
                clitic_markers=align_config.clitic_markers,
                compound_markers=align_config.compound_markers,
                multilingual_ipa=acoustic_model.meta["multilingual_ipa"],
                strip_diacritics=acoustic_model.meta.get("strip_diacritics", None),
                digraphs=acoustic_model.meta.get("digraphs", None),
            )
        else:
            dictionary = Dictionary(
                args.dictionary_path,
                data_directory,
                logger=logger,
                punctuation=align_config.punctuation,
                clitic_markers=align_config.clitic_markers,
                compound_markers=align_config.compound_markers,
                multilingual_ipa=acoustic_model.meta["multilingual_ipa"],
                strip_diacritics=acoustic_model.meta.get("strip_diacritics", None),
                digraphs=acoustic_model.meta.get("digraphs", None),
            )
        acoustic_model.validate(dictionary)

        begin = time.time()
        previous = PretrainedAligner(
            corpus,
            dictionary,
            acoustic_model,
            align_config,
            temp_directory=data_directory,
            debug=debug,
            logger=logger,
        )
        if args.full_train:
            training_config = acoustic_model.adaptation_config()
            training_config.training_configs[0].update(
                {"beam": align_config.beam, "retry_beam": align_config.retry_beam}
            )
            training_config.update_from_align(align_config)
            logger.debug("ADAPT TRAINING CONFIG:")
            log_config(logger, training_config)
            a = TrainableAligner(
                corpus,
                dictionary,
                training_config,
                align_config,
                temp_directory=data_directory,
                debug=debug,
                logger=logger,
                pretrained_aligner=previous,
            )
            logger.debug(f"Setup adapter trainer in {time.time() - begin} seconds")
            a.verbose = args.verbose
            generate_final_alignments = True
            if args.output_directory is None:
                generate_final_alignments = False
            else:
                os.makedirs(args.output_directory, exist_ok=True)

            begin = time.time()
            a.train(generate_final_alignments)
            logger.debug(f"Trained adapted model in {time.time() - begin} seconds")
            if args.output_model_path is not None:
                a.save(args.output_model_path, root_directory=model_directory)

            if generate_final_alignments:
                a.export_textgrids(args.output_directory)

            a.save(args.output_model_path, root_directory=model_directory)
        else:
            a = AdaptingAligner(
                corpus,
                dictionary,
                previous,
                align_config,
                temp_directory=data_directory,
                debug=debug,
                logger=logger,
            )
            logger.debug(f"Setup adapter trainer in {time.time() - begin} seconds")
            a.verbose = args.verbose
            generate_final_alignments = True
            if args.output_directory is None:
                generate_final_alignments = False
            else:
                os.makedirs(args.output_directory, exist_ok=True)
            begin = time.time()
            a.adapt()
            logger.debug(f"Mapped adapted model in {time.time() - begin} seconds")
            if args.output_model_path is not None:
                a.save(args.output_model_path, root_directory=model_directory)
            if generate_final_alignments:
                begin = time.time()
                a.align()
                logger.debug(
                    f"Generated alignments with adapted model in {time.time() - begin} seconds"
                )

            if generate_final_alignments:
                a.export_textgrids(args.output_directory)

            a.save(args.output_model_path, root_directory=model_directory)

        logger.info("All done!")
        logger.debug(f"Done! Everything took {time.time() - all_begin} seconds")
    except Exception:
        conf["dirty"] = True
        raise
    finally:
        handlers = logger.handlers[:]
        for handler in handlers:
            handler.close()
            logger.removeHandler(handler)
        conf.save(conf_path)


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    try:
        args.speaker_characters = int(args.speaker_characters)
    except ValueError:
        pass

    args.output_directory = None
    if not args.output_model_path:
        args.output_model_path = None
    output_paths = args.output_paths
    if len(output_paths) > 2:
        raise ArgumentError(f"Got more arguments for output_paths than 2: {output_paths}")
    for path in output_paths:
        if path.endswith(".zip"):
            args.output_model_path = path
        else:
            args.output_directory = path.rstrip("/").rstrip("\\")

    args.corpus_directory = args.corpus_directory.rstrip("/").rstrip("\\")
    if not os.path.exists(args.corpus_directory):
        raise ArgumentError(f"Could not find the corpus directory {args.corpus_directory}.")
    if not os.path.isdir(args.corpus_directory):
        raise ArgumentError(
            f"The specified corpus directory ({args.corpus_directory}) is not a directory."
        )

    args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")
    args.acoustic_model_path = validate_model_arg(args.acoustic_model_path, "acoustic")


def run_adapt_model(args: Namespace, unknown_args: Optional[Collection] = None) -> None:
    """
    Wrapper function for running acoustic model adaptation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: List[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    adapt_model(args, unknown_args)



"""Class definitions for the MFA transcriber"""
from __future__ import annotations

import multiprocessing as mp
import os
import shutil
import subprocess
from typing import TYPE_CHECKING, Optional, Tuple

from .config import TEMP_DIR
from .exceptions import KaldiProcessingError
from .helper import score
from .multiprocessing.transcription import (
    create_hclgs,
    score_transcriptions,
    transcribe,
    transcribe_fmllr,
)
from .utils import log_kaldi_errors, thirdparty_binary

if TYPE_CHECKING:
    from logging import Logger

    from .config.transcribe_config import TranscribeConfig
    from .corpus import TranscribeCorpus
    from .dictionary import DictionaryType
    from .models import AcousticModel, LanguageModel

__all__ = ["Transcriber"]


class Transcriber:
    """
    Class for performing transcription.

    Parameters
    ----------
    corpus : :class:`~montreal_forced_aligner.corpus.base.Corpus`
        Corpus to transcribe
    dictionary: :class:`~montreal_forced_aligner.dictionary.Dictionary`
        Pronunciation dictionary to use as a lexicon
    acoustic_model : :class:`~montreal_forced_aligner.models.AcousticModel`
        Acoustic model to use
    language_model : :class:`~montreal_forced_aligner.models.LanguageModel`
        Language model to use
    transcribe_config : :class:`~montreal_forced_aligner.config.TranscribeConfig`
        Language model to use
    temp_directory : str, optional
        Specifies the temporary directory root to save files need for Kaldi.
        If not specified, it will be set to ``~/Documents/MFA``
    debug : bool
        Flag for running in debug mode, defaults to false
    verbose : bool
        Flag for running in verbose mode, defaults to false
    evaluation_mode : bool
        Flag for running in evaluation mode, defaults to false
    logger : :class:`~logging.Logger`, optional
        Logger to use
    """

    min_language_model_weight = 7
    max_language_model_weight = 17
    word_insertion_penalties = [0, 0.5, 1.0]

    def __init__(
        self,
        corpus: TranscribeCorpus,
        dictionary: DictionaryType,
        acoustic_model: AcousticModel,
        language_model: LanguageModel,
        transcribe_config: TranscribeConfig,
        temp_directory: Optional[str] = None,
        debug: bool = False,
        verbose: bool = False,
        evaluation_mode: bool = False,
        logger: Optional[Logger] = None,
    ):
        self.logger = logger
        self.corpus = corpus
        self.dictionary = dictionary
        self.acoustic_model = acoustic_model
        self.language_model = language_model
        self.transcribe_config = transcribe_config

        if not temp_directory:
            temp_directory = TEMP_DIR
        self.temp_directory = temp_directory
        self.verbose = verbose
        self.debug = debug
        self.evaluation_mode = evaluation_mode
        self.acoustic_model.export_model(self.model_directory)
        self.acoustic_model.export_model(self.working_directory)
        self.log_dir = os.path.join(self.transcribe_directory, "log")
        self.uses_voiced = False
        self.uses_splices = False
        self.uses_cmvn = True
        self.speaker_independent = True
        os.makedirs(self.log_dir, exist_ok=True)
        self.setup()

    @property
    def transcribe_directory(self) -> str:
        """Temporary directory root for all transcription"""
        return os.path.join(self.temp_directory, "transcribe")

    @property
    def evaluation_directory(self):
        """Evaluation directory path for the current language model weight and word insertion penalty"""
        eval_string = f"eval_{self.transcribe_config.language_model_weight}_{self.transcribe_config.word_insertion_penalty}"
        path = os.path.join(self.working_directory, eval_string)
        os.makedirs(path, exist_ok=True)
        return path

    @property
    def working_directory(self) -> str:
        """Current working directory"""
        return self.transcribe_directory

    @property
    def evaluation_log_directory(self) -> str:
        """Log directory for the current evaluation"""
        return os.path.join(self.evaluation_directory, "log")

    @property
    def working_log_directory(self) -> str:
        """Log directory for the current state"""
        return os.path.join(self.working_directory, "log")

    @property
    def data_directory(self) -> str:
        """Corpus data directory"""
        return self.corpus.split_directory

    @property
    def model_directory(self) -> str:
        """Model directory for the transcriber"""
        return os.path.join(self.temp_directory, "models")

    @property
    def model_path(self) -> str:
        """Acoustic model file path"""
        return os.path.join(self.working_directory, "final.mdl")

    @property
    def alignment_model_path(self) -> str:
        """Alignment (speaker-independent) acoustic model file path"""
        path = os.path.join(self.working_directory, "final.alimdl")
        if os.path.exists(path):
            return path
        return self.model_path

    @property
    def fmllr_options(self):
        """Options for computing fMLLR transforms"""
        data = self.transcribe_config.fmllr_options
        data["sil_phones"] = self.dictionary.silence_csl
        return data

    @property
    def hclg_options(self):
        """Options for constructing HCLG FSTs"""
        context_width, central_pos = self.get_tree_info()
        return {
            "context_width": context_width,
            "central_pos": central_pos,
            "self_loop_scale": self.transcribe_config.self_loop_scale,
            "transition_scale": self.transcribe_config.transition_scale,
        }

    def get_tree_info(self) -> Tuple[int, int]:
        """
        Get the context width and central position for the acoustic model

        Returns
        -------
        int
            Context width
        int
            Central position
        """
        tree_proc = subprocess.Popen(
            [thirdparty_binary("tree-info"), os.path.join(self.model_directory, "tree")],
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, _ = tree_proc.communicate()
        context_width = 1
        central_pos = 0
        for line in stdout.split("\n"):
            text = line.strip().split(" ")
            if text[0] == "context-width":
                context_width = int(text[1])
            elif text[0] == "central-position":
                central_pos = int(text[1])
        return context_width, central_pos

    def setup(self) -> None:
        """
        Sets up the corpus and transcriber

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        dirty_path = os.path.join(self.model_directory, "dirty")

        if os.path.exists(dirty_path):  # if there was an error, let's redo from scratch
            shutil.rmtree(self.model_directory)
        log_dir = os.path.join(self.model_directory, "log")
        os.makedirs(log_dir, exist_ok=True)
        self.dictionary.write(write_disambiguation=True)
        for dict_name, output_directory in self.dictionary.output_paths.items():
            words_path = os.path.join(self.model_directory, f"words.{dict_name}.txt")
            shutil.copyfile(os.path.join(output_directory, "words.txt"), words_path)
        self.corpus.initialize_corpus(self.dictionary, self.acoustic_model.feature_config)

        big_arpa_path = self.language_model.carpa_path
        small_arpa_path = self.language_model.small_arpa_path
        medium_arpa_path = self.language_model.medium_arpa_path
        if not os.path.exists(small_arpa_path) or not os.path.exists(medium_arpa_path):
            self.logger.info("Parsing large ngram model...")
            mod_path = os.path.join(self.model_directory, "base_lm.mod")
            new_carpa_path = os.path.join(self.model_directory, "base_lm.arpa")
            with open(big_arpa_path, "r", encoding="utf8") as inf, open(
                new_carpa_path, "w", encoding="utf8"
            ) as outf:
                for line in inf:
                    outf.write(line.lower())
            big_arpa_path = new_carpa_path
            subprocess.call(["ngramread", "--ARPA", big_arpa_path, mod_path])

            if not os.path.exists(small_arpa_path):
                self.logger.info(
                    "Generating small model from the large ARPA with a pruning threshold of 3e-7"
                )
                prune_thresh_small = 0.0000003
                small_mod_path = mod_path.replace(".mod", "_small.mod")
                subprocess.call(
                    [
                        "ngramshrink",
                        "--method=relative_entropy",
                        f"--theta={prune_thresh_small}",
                        mod_path,
                        small_mod_path,
                    ]
                )
                subprocess.call(["ngramprint", "--ARPA", small_mod_path, small_arpa_path])

            if not os.path.exists(medium_arpa_path):
                self.logger.info(
                    "Generating medium model from the large ARPA with a pruning threshold of 1e-7"
                )
                prune_thresh_medium = 0.0000001
                med_mod_path = mod_path.replace(".mod", "_med.mod")
                subprocess.call(
                    [
                        "ngramshrink",
                        "--method=relative_entropy",
                        f"--theta={prune_thresh_medium}",
                        mod_path,
                        med_mod_path,
                    ]
                )
                subprocess.call(["ngramprint", "--ARPA", med_mod_path, medium_arpa_path])
        try:
            create_hclgs(self)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def transcribe(self) -> None:
        """
        Transcribe the corpus

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self.logger.info("Beginning transcription...")
        dirty_path = os.path.join(self.transcribe_directory, "dirty")
        if os.path.exists(dirty_path):
            shutil.rmtree(self.transcribe_directory, ignore_errors=True)
        os.makedirs(self.log_dir, exist_ok=True)
        try:
            transcribe(self)
            if (
                self.acoustic_model.feature_config.fmllr
                and not self.transcribe_config.ignore_speakers
                and self.transcribe_config.fmllr
            ):
                self.logger.info("Performing speaker adjusted transcription...")
                transcribe_fmllr(self)
            score_transcriptions(self)
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger.handlers[0].baseFilename)
            raise

    def evaluate(self):
        """
        Evaluates the transcripts if there are reference transcripts

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        self.logger.info("Evaluating transcripts...")
        self._load_transcripts()
        # Sentence-level measures

        correct = 0
        incorrect = 0
        # Word-level measures
        total_edits = 0
        total_length = 0
        issues = []
        with mp.Pool(self.corpus.num_jobs) as pool:
            to_comp = []
            for utt_name, utterance in self.corpus.utterances.items():
                g = utterance.text.split()
                if not utterance.transcription_text:
                    incorrect += 1
                    total_edits += len(g)
                    total_length += len(g)

                h = utterance.transcription_text.split()
                if g != h:
                    issues.append((utt_name, g, h))
                to_comp.append((g, h))
            gen = pool.starmap(score, to_comp)
            for (edits, length) in gen:
                if edits == 0:
                    correct += 1
                else:
                    incorrect += 1
                total_edits += edits
                total_length += length
        ser = 100 * incorrect / (correct + incorrect)
        wer = 100 * total_edits / total_length
        output_path = os.path.join(self.evaluation_directory, "transcription_issues.csv")
        with open(output_path, "w", encoding="utf8") as f:
            for utt, g, h in issues:
                g = " ".join(g)
                h = " ".join(h)
                f.write(f"{utt},{g},{h}\n")
        self.logger.info(f"SER: {ser:.2f}%, WER: {wer:.2f}%")
        return ser, wer

    def _load_transcripts(self):
        """Load transcripts from Kaldi temporary files"""
        for j in self.corpus.jobs:
            score_arguments = j.score_arguments(self)
            for tra_path in score_arguments.tra_paths.values():

                with open(tra_path, "r", encoding="utf8") as f:
                    for line in f:
                        t = line.strip().split(" ")
                        utt = t[0]
                        utterance = self.corpus.utterances[utt]
                        speaker = utterance.speaker
                        lookup = speaker.dictionary.reversed_word_mapping
                        ints = t[1:]
                        if not ints:
                            continue
                        transcription = []
                        for i in ints:
                            transcription.append(lookup[int(i)])
                        utterance.transcription_text = " ".join(transcription)

    def export_transcriptions(self, output_directory: str) -> None:
        """
        Export transcriptions

        Parameters
        ----------
        output_directory: str
            Directory to save transcriptions
        """
        backup_output_directory = None
        if not self.transcribe_config.overwrite:
            backup_output_directory = os.path.join(self.transcribe_directory, "transcriptions")
            os.makedirs(backup_output_directory, exist_ok=True)
        self._load_transcripts()
        for file in self.corpus.files.values():
            file.save(output_directory, backup_output_directory)
